#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "types.h"
#include "header.h"
#include "key.h"
#include "icon.h"
#include "menu.h"
#include "files.h"

int wscreen,hscreen, wdesk, hdesk, xdesk, ydesk; /* size of the window */

Display *display;
char sdisp[64]="";
char *backcolor="white";
int screen, depth;
Window window,twindow,bwindow,vwindow,vuwindow,vdwindow,swindow,
	awindow,a1window,aswindow[4];
GC gc,invgc,xorgc;
unsigned long blackpixel, whitepixel;
long memsize=256*1024l; /* size of memory to be used */

XFontStruct *textfont;
int ascent,descent;

char textfontname[256]=DEFFONT;
int userwidth=800,userheight=600,userx=0,usery=0,usersize=0,userpos=0;
/* those are the default things, may be set by -geom ... */

int gl_hchar,gl_wchar,textheight,textwidth,textoffset,voffset,hoffset;
int twidth,theight,vwidth,vheight,bwidth,bheight;

void init_font (void)
{
	/* load a font */
	textfont=XLoadQueryFont(display,textfontname);
	if (!textfont)
	{
		fprintf(stderr,"Cannot find %s font\n",textfontname);
		exit(1);
	}
	gl_hchar=textheight=textfont->ascent+textfont->descent;
	gl_wchar=textwidth=XTextWidth(textfont,"m",1);
	ascent=textfont->ascent;
	descent=textfont->descent;
	if (textwidth!=XTextWidth(textfont,"i",1))
	{
		fprintf(stderr,
			"You cannot use the proportional font %s!\n",textfontname);
		exit(1);
	}
	textoffset=textwidth/2;
}

void initX (void)
/* Sets up the connection to the X server, loads the text font. */
{	
	display=XOpenDisplay(sdisp);
	if (!display)
	{       
		fprintf(stderr,"ERROR: Connetion to X server %s failed.\n",
		    sdisp);
		exit(1);
	}
	screen=DefaultScreen(display); /* screen # */
	depth=DefaultDepth(display,screen); /* depth of that screen */
	init_font();
	
	XFlush(display);
}

void quitX (void)
/* Disconnect to server, free resources. */
{   
	static int quitted=0;
	if (quitted) return;
	XDestroyWindow(display,window);
	if (textfont) XFreeFont(display,textfont);
	XCloseDisplay(display);
}

int createGC (Window window, GC *gc)
/* Create a new graphic context */
{	
	XGCValues gcvalues;
	*gc=XCreateGC(display,window,(unsigned long)0,&gcvalues);
	if (*gc==0) exit(1);
	return 1;
}

void makegc (Window w, GC *gc, unsigned long forg, unsigned long backg)
{
	createGC(w,gc);
	XSetForeground(display,*gc,forg);
	XSetBackground(display,*gc,backg);
	XSetFont(display,*gc,textfont->fid);
}

void open_window (int x, int y, int width, int height)
/* Create and open a window, set window name and class name, set
icon name, give size hints, create contexts, choose input,
map the window. */
{	
	XSetWindowAttributes attributes;
	XSizeHints sizehints;
	Pixmap iconpixmap;
	XWMHints wmhints;
	/* XClassHint classhints; */
	unsigned int cursorshape;
	Cursor cursor;
	unsigned long black,white;
	unsigned long mask;
	int i;
	XColor rgb,hardware;
	static Colormap colormap;
		
	/* window attributes */
	black=attributes.border_pixel=BlackPixel(display,screen);
	
	colormap=DefaultColormap(display,screen);
	if (XLookupColor(display,colormap,backcolor,&rgb,&hardware))
		if (XAllocColor(display,colormap,&hardware))
			white=hardware.pixel;
/*	white=attributes.background_pixel=WhitePixel(display,screen); */

	mask= CWBackPixel | CWBorderPixel;

	/* create the window */
	window= XCreateWindow(display,RootWindow(display,screen),
	    x,y,width,height,
	    2, /* Border width */
	    depth,
	    InputOutput,
	    CopyFromParent,
	    mask,
	    &attributes);

	/* class hints
	classhints.res_name="editor";
	classhints.res_class="Editor";
	XSetClassHint(display,window,&classhints); */

	/* Window and Icon name */
	XStoreName(display,window,"Editor");
	XSetIconName(display,window,"e");

	iconpixmap=XCreateBitmapFromData(display,window,(char *)icon_bits,
		icon_width,icon_height);
	wmhints.icon_pixmap=iconpixmap;
	wmhints.initial_state=NormalState;
	/* Setting the input hint for some window managers */
    wmhints.input = True;
    wmhints.flags = IconPixmapHint|StateHint|InputHint;
    XSetWMHints(display, window, &wmhints);

	/* size hints */
	sizehints.flags=PMinSize|PResizeInc;
	if (userpos) sizehints.flags |= USPosition | USSize;
	else sizehints.flags |= PPosition | USSize;
	sizehints.x=x;
	sizehints.y=y;
	sizehints.width=width;
	sizehints.height=height;
	sizehints.min_width=60*gl_wchar+2*hoffset+6;
	sizehints.min_height=25*gl_hchar+4*hoffset+6;
	sizehints.width_inc=gl_wchar;
	sizehints.height_inc=gl_hchar;
	XSetNormalHints(display,window,&sizehints);

	/* Cursor */
	cursorshape=XC_hand2;
	cursor=XCreateFontCursor(display,cursorshape);
	XDefineCursor(display,window,cursor);

	/* events, we like to receive */
	XSelectInput(display,window,StructureNotifyMask|KeyPressMask);

	makegc(window,&gc,black,white);
	makegc(window,&invgc,white,black);
#ifdef BLACK0
	makegc(window,&xorgc,white,black);
#else
	makegc(window,&xorgc,black,white);
#endif
	XSetFunction(display,xorgc,GXxor);

	twidth=width-gl_wchar*3-6;
	theight=height-2*hoffset-2-gl_hchar-6;
	twindow=XCreateSimpleWindow(display,window,
		0,0,twidth,theight,
		2,black,
		white);
	XSelectInput(display,twindow,ExposureMask|ButtonPressMask);

	bwindow=XCreateSimpleWindow(display,window,
		0,theight+4,bwidth=twidth+2,bheight=height-theight-6,
		1,black,white);
	XSelectInput(display,bwindow,ExposureMask|ButtonPressMask);

	vwidth=width-twidth-6;
	vheight=height-2*vwidth-6;
	vwindow=XCreateSimpleWindow(display,window,
		twidth+4,vwidth,vwidth+2,vheight,
		1,black,white);
	XSelectInput(display,vwindow,ButtonPressMask|ButtonReleaseMask);

	vuwindow=XCreateSimpleWindow(display,window,
		twidth+4,0,vwidth,vwidth,
		1,black,white);
	XSelectInput(display,vuwindow,ExposureMask|ButtonPressMask|
		ButtonReleaseMask);

	vdwindow=XCreateSimpleWindow(display,window,
		twidth+4,height-vwidth-2,vwidth,vwidth,
		1,black,white);
	XSelectInput(display,vdwindow,ExposureMask|ButtonPressMask|
		ButtonReleaseMask);

	swindow=XCreateSimpleWindow(display,vwindow,
		2,2,vwidth-8,vwidth-8,
		2,black,white);
	XSelectInput(display,swindow,ButtonPressMask|ButtonReleaseMask);

	awindow=XCreateSimpleWindow(display,twindow,
		0,0,10,10,2,black,white);
	XSelectInput(display,awindow,ButtonPressMask|ExposureMask);
	for (i=0; i<4; i++)
	{	aswindow[i]=XCreateSimpleWindow(display,awindow,
			0,0,10,10,1,black,white);
		XSelectInput(display,aswindow[i],ButtonPressMask);
	}

	/* show the window */
	XMapWindow(display,window);
	XMapSubwindows(display,window);
	XMapSubwindows(display,vwindow);

	XFlush(display);
}

void grafik (void)
/* switch to a graphic screen, i.e. open the window, if not already
done, connect with server. */
{
	static int winit=0;
	if (!winit)
	{	
		initX();
		atexit(quitX);
		if (usersize)
		{	
			userwidth=abs(userwidth);
			if (userwidth<256) userwidth=256;
			if (userwidth>DisplayWidth(display,screen))
				userwidth=DisplayWidth(display,screen);
			userheight=abs(userheight);
			if (userheight<256) userheight=256;
			if (userheight>DisplayHeight(display,screen))
				userheight=DisplayHeight(display,screen);
		}
		else
		{
			userwidth=DisplayWidth(display,screen)/3*2;
			userheight=DisplayHeight(display,screen)/3*2;
		}
		if (userpos)
		{
			if (userx<0) userx=DisplayWidth(display,screen)
				+userx-userwidth;
			if (usery<0) usery=DisplayHeight(display,screen)
				+usery-userheight;
			if (userx<0) userx=0;
			if (usery<0) usery=0;
		}
		else
		{
			userx=usery=0;
		}
		open_window(userx,usery,userwidth,userheight);
		wdesk=wscreen=userwidth;
		hdesk=hscreen=userheight;
		hoffset=gl_wchar/3; voffset=gl_hchar/3;
		xdesk=ydesk=0;
		winit=1;
	}
}

void clear_screen (void)
{
	XFillRectangle(display,twindow,invgc,0,0,twidth,theight);
	bline(0,"");
}

void process_event (XEvent *event)
{
	XEvent dummyevent;
	int key,i,p[4];
	XRectangle rect[2];
	switch (event->type)
	{
		case GraphicsExpose :
			p[0]=rect[0].x=event->xgraphicsexpose.x;
			p[1]=rect[0].y=event->xgraphicsexpose.y;
			p[2]=rect[0].width=event->xgraphicsexpose.width;
			p[3]=rect[0].height=event->xgraphicsexpose.height;
			p[2]+=p[0]; p[3]+=p[1];
			goto expose;
		case Expose :
			p[0]=rect[0].x=event->xexpose.x;
			p[1]=rect[0].y=event->xexpose.y;
			p[2]=rect[0].width=event->xexpose.width;
			p[3]=rect[0].height=event->xexpose.height;
			p[2]+=p[0]; p[3]+=p[1];
			expose:
			XSetClipRectangles(display,gc,0,0,rect,1,Unsorted);
			XSetClipRectangles(display,invgc,0,0,rect,1,Unsorted);
			XSetClipRectangles(display,xorgc,0,0,rect,1,Unsorted);
			wredraw(topslot,p,event->xexpose.window);
			XSetClipMask(display,gc,None);
			XSetClipMask(display,invgc,None);
			XSetClipMask(display,xorgc,None);
			break;
		case ConfigureNotify :
			if (event->xconfigure.width==wscreen &&
			    event->xconfigure.height==hscreen) break;
			wscreen=event->xconfigure.width;
			hscreen=event->xconfigure.height;
			twidth=wscreen-gl_wchar*3-6;
			theight=hscreen-2*hoffset-2-gl_hchar-6;
			XMoveResizeWindow(display,twindow,
				0,0,twidth,theight);
			XMoveResizeWindow(display,bwindow,
				0,theight+4,bwidth=twidth+2,bheight=hscreen-theight-6);
			vwidth=wscreen-twidth-6;
			vheight=hscreen-6-2*vwidth;
			XMoveResizeWindow(display,vwindow,
				twidth+4,vwidth+2,vwidth,vheight);
			XMoveResizeWindow(display,vuwindow,
				twidth+4,0,vwidth,vwidth);
			XMoveResizeWindow(display,vdwindow,
				twidth+4,hscreen-2-vwidth,vwidth,vwidth);
			XMoveResizeWindow(display,swindow,
				2,2,vwidth-8,vwidth-8);
			for (i=0; i<MAXSLOTS; i++)
				if (files[i].flags.used)
					wsize_file(i,0,0,twidth,theight);
			while (XCheckMaskEvent(display,ExposureMask,&dummyevent))
				process_event(&dummyevent);
			break;
		case KeyPress :
			key=translate(&event->xkey);
			if (!key) break;
			key_interpret(key);
			break;
		case ButtonPress :
			button_interpret(&event->xbutton);
			break;
		case MappingNotify :
			XRefreshKeyboardMapping(&event->xmapping);
			break;
	}
}

void usage (void)
{
	fprintf(stderr,"editor [-f(ont)Name] [-s(ize)KByte]\n"
		" [+|-i(ndent)] [+|-d(upsave)] [+|-c(onfig save)]\n"
		" [-|+w(ordwrap)] [-|+b(lanks compress)] [-|+v(ertical scroll bar)]\n"
		" [-x backgroundcolorname]\n"
		" [-t(ab)nnn] [-l(linelength)nnn] [-disp DISPLAY] [-k keysfile]\n"
		" [-geom GEOMETRY] [+position] file [+position] file ...\n");
	fprintf(stderr,"example :\n"
		" editor -f8x13 -s256 +w +b -t4 -l80"
		" =500x500+0+0 +25 file1.txt file2.txt\n");		
}

void get_geometry (char *s)
{
	long d;
	char *end;
	if (*s=='=') s++;
	d=strtol(s,&end,10);
	if (!end || end==s) return; else { userwidth=d; s=end; }
	if (*s++!='x') { usage(); exit(1); }
	d=strtol(s,&end,10);
	if (!end || end==s) return; else { userheight=d; s=end; }
	usersize=1;
	if (*s==0) return;
	d=strtol(s,&end,10);
	if (!end || end==s) return; 
	else 
	{ 
		userx=d;
		if (*s=='-' && d==0) userx=-1;
		s=end; 
	}
	userpos=1;
	if (*s==0) return;
	d=strtol(s,&end,10);
	if (!end || end==s) return;
	else 
	{ 
		usery=d;
		if (*s=='-' && d==0) usery=-1;
		s=end; 
	}
}

void errorsave (void)
{	int i;
	static int done=0;
	char name[512];
	FILE *file;
	if (done) return;
	for (i=0; i<MAXSLOTS; i++)
		if (files[i].flags.used && files[i].flags.updated)
		{	close_gap(i);
			strcpy(name,files[i].name);
			strcat(name,".save");
			file=fopen(name,"w");
			if (file)
			{	fwrite(files[i].start,(size_t)1,
					files[i].fend-files[i].start,file);
				fclose(file);
			}
		}
	done=1;
}

void backsave (void)
{	int i;
	char name[512];
	FILE *file;
	for (i=0; i<MAXSLOTS; i++)
		if (files[i].flags.used && files[i].flags.updated)
		{	close_gap(i);
			strcpy(name,files[i].name);
			strcat(name,".save");
			file=fopen(name,"w");
			if (file)
			{	fwrite(files[i].start,(size_t)1,
					files[i].fend-files[i].start,file);
				fclose(file);
			}
		}
}

void errorhandler (Display *display)
{	errorsave();
	exit(1);
}

void ioerrorhandler (Display *display)
{	errorsave();
	exit(1)	;
}

void insterror (void)
{	XSetErrorHandler((XErrorHandler)errorhandler);
	XSetIOErrorHandler((XIOErrorHandler)ioerrorhandler);
}

extern char * translatet[256];

void load_translations (char *filename)
{	FILE *in;
	int c,cc;
	char s[256],*p;
	in=fopen(filename,"r");
	if (!in) return;
	while (!feof(in))
	{	cc=fgetc(in);
		if (cc==EOF) break;
		p=s;
		while (1)
		{	c=fgetc(in);
			if (c==EOF || c=='\n') break;
			if (p-s<255) *p++=c;
		}
		*p=0;
		if (s[0])
		{	p=(char *)malloc(strlen(s)+1);
			strcpy(p,s);
			translatet[cc]=p;
		}
	}
	fclose(in);
	dotranslate=1;
}

int main (int argc, char *argv[])
{
	char *p;
	XEvent event;
	while (argc>1)
	{
		if (argv[1][0]=='=')
		{	get_geometry(argv[1]+1);
			argc--; argv++;
		}
		else if (argv[1][0]=='-')
		{
			switch(argv[1][1])
			{
				case 'F' :
				case 'f' :
					if (argv[1][2]) strcpy(textfontname,argv[1]+2);
					else if (argc>2)
					{	strcpy(textfontname,argv[2]); argv++; argc--;
					}
					break;
				case 'S' :
				case 's' :
					if (argv[1][2]) memsize=atoi(argv[1]+2)*1024l;
					else if (argc>2)
					{	memsize=atoi(argv[2])*1024l; argv++; argc--; }
					if (memsize<10*1024l) memsize=10*1024l;
					break;
				case 'K' :
				case 'k' :
					if (argv[1][2]) p=argv[1]+2;
					else if (argc>2) { p=argv[2]; argv++; argc--; }
					else break;
					load_translations(p);
					break;
				case 'I' :
				case 'i' : params.indent=0; break;
				case 'B' :
				case 'b' : params.blanks=0; break;
				case 'W' :
				case 'w' : params.wordwrap=0; break;
				case 'D' :
				case 'd' :
					if (strcmp(argv[1],"-disp")==0)
					{	strcpy(sdisp,argv[2]); argc--; argv++;
					}
					else params.dup=0;
					break;
				case 'C' :
				case 'c' : params.config=0; break;
				case 'T' :
				case 't' :
					if (argv[1][2])
						params.tabsize=atoi(argv[1]+2);
					else
					{	params.tabsize=atoi(argv[2]);
						argv++; argc--;
					}
					if (params.tabsize>16 || params.tabsize<0)
						params.tabsize=4;
					break;
				case 'L' :
				case 'l' :
					if (argv[1][2])
						params.linelength=atoi(argv[1]+2);
					else
					{	params.linelength=atoi(argv[2]);
						argv++; argc--;
					}
					if (params.linelength<32)
						params.linelength=32;
					break;
				case 'v' :
				case 'V' : params.scrollbar=0; break;
				case 'g' :
				case 'G' : if (argc<=2) break;
					get_geometry(argv[2]); argc--; argv++; break;
				case 'x' :
				case 'X' : if (argc<=2) break;
					backcolor=argv[2]; argc--; argv++; break;
				default : usage(); exit(1);
			}
			argc--; argv++;
		}
		else if (argv[1][0]=='+')
		{
			switch(argv[1][1])
			{
				case 'I' :
				case 'i' : params.indent=1; break;
				case 'B' :
				case 'b' : params.blanks=1; break;
				case 'W' :
				case 'w' : params.wordwrap=1; break;
				case 'D' :
				case 'd' : params.dup=1; break;
				case 'C' :
				case 'c' : params.config=1; break;
				case 'v' :
				case 'V' : params.scrollbar=1; break;
				default : 
					if (argv[1][2]<'0' || argv[1][2]>'9')
					{	usage(); exit(1);
					}
					goto cont;
			}
			argc--; argv++;
		}
		else break;
	}
	cont:
	grafik();
	insterror();
	do 
	{	XWindowEvent(display,window,StructureNotifyMask,&event);
		process_event(&event);
	}
	while (event.type!=MapNotify);
	main_loop(argc,argv);
	return 0;
}
